/*------------------------------------------------------------------------
 Module:        output_ogg.c
 Author:        (c) 2003 Felix de las Pozas Alvarez
 Project:       Soul Reaver 1 music extractor and converter
 State:         v1.1 FINAL VERSION
 Creation Date: December 2 2003
 Description:   Write the files in OGG Vorbis format. Source file.
 License:       GNU Public License 
 
    This file is part of SR_Music.
    SR_Music free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    SR_Music is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with SHA-1; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
    USA
------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "ogg/include/vorbis/vorbisenc.h"
#include "ogg/include/ogg/ogg.h"

void ogg_output(FILE *musicdat)
{
    /* songs positions in music.dat */
    const unsigned long POSITION[] =
        {
            0,
            9838576,
            23828720,
            33274480,
            43545200,
            53122640,
            63134608,
            76862256,
            92508192,
            101179808,
            111809812
        };

    FILE *output, *playlist;
    unsigned char output_name[120];
    unsigned char name_temp[120];
    signed char   ogg_buffer[4096];
    unsigned char result;
    unsigned char loop;
    unsigned int  loop2;
    unsigned long size;
    unsigned long coding_ended = 0;
    unsigned long bytes;
    float         **samples_buffer;

    /* ogg init stuff */
    ogg_stream_state os;
    ogg_page         og;
    ogg_packet       op;
    vorbis_info      vi;
    vorbis_comment   vc;
    vorbis_dsp_state vd;
    vorbis_block     vb;

    /* try to open playlist file, notify errors and exit */
    if (NULL == (playlist = fopen("Kurt Harland - Soul Reaver.m3u", "wb")))
    {
        printf("ERROR: Error creating playlist file.\n");
        exit(-1);
    }

    for (loop = 0; loop < 10; loop++)
    {
        /* create name & write it to the playlist */
        strcpy(output_name, "Track ");
        itoa(loop+1, name_temp, 10);
        strcat(output_name, name_temp);
        strcpy(name_temp, output_name);
        strcat(output_name, ".ogg");

        fputs(output_name, playlist);
        fputs("\n", playlist);

        /* try to open output file, notify errors and exit */
        if (NULL == (output = fopen(output_name, "wb")))
        {
            printf("ERROR: Error opening output file %s\n",output_name);
            exit(-1);
        }

        size = POSITION[loop+1] - POSITION[loop];
        fseek(musicdat, POSITION[loop], SEEK_SET);

        printf("Writting file : \"%s\" ... ", output_name);

        /* ogg encode setup */
        vorbis_info_init(&vi);

        /* set the encoding mode, notify errors */
        if (vorbis_encode_init(&vi, 2, 22050, -1, 96000, -1))
        {
            printf("\nERROR: In OGG encoding mode setup\n");
            exit(-1);
        }

        /* add comments to the song */
        vorbis_comment_init(&vc);
        vorbis_comment_add_tag(&vc,"TITLE",name_temp);
        vorbis_comment_add_tag(&vc,"ALBUM","Soul Reaver Soundtrack");
        itoa(loop+1, output_name, 10);
        vorbis_comment_add_tag(&vc,"TRACKNUMBER",output_name);
        vorbis_comment_add_tag(&vc,"ARTIST","Kurt Harland");
        vorbis_comment_add_tag(&vc,"GENRE","Soundtrack");
        vorbis_comment_add_tag(&vc,"COPYRIGHT","(C) Eidos Interactive");

        /* set up the analysis state and auxiliary encoding storage */
        vorbis_analysis_init(&vd,&vi);
        vorbis_block_init(&vd,&vb);

        /* set up our packet->stream encoder */
        srand(time(NULL));
        ogg_stream_init(&os,rand());

        /* three headers for OGG files */
        ogg_packet header;
        ogg_packet header_comm;
        ogg_packet header_code;

        vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
        ogg_stream_packetin(&os,&header);
        ogg_stream_packetin(&os,&header_comm);
        ogg_stream_packetin(&os,&header_code);

        /* write OGG header */
        while ((result = ogg_stream_flush(&os,&og)))
        {
            if (!result)
                break;

            if (og.header_len != fwrite(og.header,1,og.header_len, output))
            {
                printf("\nERROR: I/O Error writting OGG header %s\n",output_name);
                exit(-1);
            }

            if (og.body_len != fwrite(og.body,1,og.body_len, output))
            {
                printf("\nERROR: I/O Error writting OGG header %s\n",output_name);
                exit(-1);
            }
        }

        /* data encoding */
        while (!coding_ended)
        {
            /* read PCM samples in 4096 blocks if we can */
            if (size > 4096)
            {
                /* read PCM interleaved samples */
                if (4096 != (bytes = fread(ogg_buffer,1,4096,musicdat)))
                {
                    printf("\nERROR: I/O Error reading music.dat\n");
                    exit(-1);
                }
                size -= 4096;
            }
            else
            {
                if (size != (bytes = fread(ogg_buffer,1,size,musicdat)))
                {
                    printf("\nERROR: I/O Error reading music.dat\n");
                    exit(-1);
                }
                size = 0;
            }

            printf("%3d %%\b\b\b\b\b", (int) ((((POSITION[loop+1] - POSITION[loop]) - size) * 100) / (POSITION[loop+1] - POSITION[loop])));
            
            if (bytes == 0)
            {
                vorbis_analysis_wrote(&vd,0);
            }
            else
            {
                /* there are samples to encode */

                /* expose the buffer to submit data */
                samples_buffer = vorbis_analysis_buffer(&vd,1024);

                /* uninterleave samples for OGG (float format) */
                for(loop2 = 0; loop2 < bytes/4 ; loop2++)
                {
                    samples_buffer[0][loop2]=((ogg_buffer[loop2*4+1]<<8)|
                                              (0x00ff&(int)ogg_buffer[loop2*4]))/32768.f;
                    samples_buffer[1][loop2]=((ogg_buffer[loop2*4+3]<<8)|
                                              (0x00ff&(int)ogg_buffer[loop2*4+2]))/32768.f;
                }

                /* tell the library how much we actually submitted */
                vorbis_analysis_wrote(&vd,loop2);
            }

            /* vorbis does some data preanalysis, then divvies up blocks for
               more involved (potentially parallel) processing.  Get a single
               block for encoding now */
            while(vorbis_analysis_blockout(&vd,&vb) == 1)
            {
                /* analysis, assume we want to use bitrate management */
                vorbis_analysis(&vb,NULL);
                vorbis_bitrate_addblock(&vb);

                while(vorbis_bitrate_flushpacket(&vd,&op))
                {
                    /* weld the packet into the bitstream */
                    ogg_stream_packetin(&os,&op);

                    /* write out pages (if any) */
                    while ((result = ogg_stream_flush(&os,&og)))
                    {
                        if (!result)
                            break;

                        if (og.header_len != fwrite(og.header,1,og.header_len, output))
                        {
                            printf("\nERROR: I/O Error writting OGG header %s\n",output_name);
                            exit(-1);
                        }

                        if (og.body_len != fwrite(og.body,1,og.body_len, output))
                        {
                            printf("\nERROR: I/O Error writting OGG header %s\n",output_name);
                            exit(-1);
                        }

                        if ( ogg_page_eos(&og) )
                            coding_ended = 1;
                    }
                }
            }
            bytes = 0;
        }

        /* close the output file */
        if (fclose(output) == EOF)
        {
            printf("\nERROR: Error closing output file %s\n", output_name);
        }
        printf("OK!  \n");

        /* clean up */
        ogg_stream_clear(&os);
        vorbis_block_clear(&vb);
        vorbis_dsp_clear(&vd);
        vorbis_comment_clear(&vc);
        vorbis_info_clear(&vi);
        coding_ended = 0;
    }

    printf("Writting M3U playlist ... OK!\n");

    /* close the playlist ile */
    if (fclose(playlist) == EOF)
    {
        printf("ERROR: Error closing playlist file.");
    }

    return;
}
