/*------------------------------------------------------------------------
 Module:        sha-1.c
 Author:        (c) 2003 Felix de las Pozas Alvarez
 Project:       SHA-1
 State:         v1.3 Final version
 Creation Date: December 2 2003
 Description:   Calculates the "Secure Hash Algorithm SHA-1"
                of (a) given file(s). Based on the algorithm
                described in "The Handbook Of Applied
                Cryptography" and the Federal Information
                Processing Standarts publication of august 1
                2002. Source file.
 License:       GNU Public License 
 
    This file is part of SHA-1.
    SHA-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    SHA-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with SHA-1; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
    USA
------------------------------------------------------------------------*/

#include <malloc.h>
#include "sha-1.h"

/* Must be 64*N */
#define BUFFER_SIZE 1280000

/* Initial values */
const unsigned IV[] =
    {
        0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476, 0xC3D2E1F0
    };

/* Per round integer aditive constants */
const unsigned long PRIAC[] =
    {
        0x5A827999, 0x6ED9EBA1, 0x8F1BBCDC, 0xCA62C1D6
    };

/* Rotational shift to the left */
#define ROTL( x, n ) ( (x << n) | ( ( x & 0xFFFFFFFF ) >> ( 32 - n ) ) )

/* Initialize the process */
void sha_1_init(void)
{
    unsigned char loop;

    /* initialize chaining variables (intermediate hash value) */
    for (loop = 0; loop < 5; loop++)
        chaining_vars[loop] = IV[loop];

    return;
}

/* Different functions depending of the loop index */
unsigned long Function(unsigned char loop, unsigned long x,
                       unsigned long y, unsigned long z)
{
    if (loop < 20)
        return ( ( z ^ ( x & ( y ^ z ) ) ) + PRIAC[0] );
    else
        if (loop < 40)
            return ( ( x ^ y ^ z ) + PRIAC[1] );
        else
            if (loop < 60)
                return ( ( (x & y) | (z & (x | y) ) ) + PRIAC[2]);
    return ( ( x ^ y ^ z ) + PRIAC[3]);
}

/* Process a single block */
void sha_1_process_block(unsigned char *char_block)
{

    unsigned long a, b, c, d, e, t;
    unsigned long expanded_blk[80];
    unsigned char loop;

    /* convert the block from unsigned char to unsigned long */
    for (loop = 0; loop < 16; loop++)
        expanded_blk[loop] =
            ((unsigned long) (char_block[(loop*4)]   << 24)) |
            ((unsigned long) (char_block[(loop*4)+1] << 16)) |
            ((unsigned long) (char_block[(loop*4)+2] <<  8)) |
            ((unsigned long) (char_block[(loop*4)+3]));

    /* expanding the block from 16 to 80 */
    for (loop = 16; loop < 80; loop++)
        expanded_blk[loop] = ROTL((((expanded_blk[loop-3]   ^
                                     expanded_blk[loop-8])  ^
                                    expanded_blk[loop-14]) ^
                                   expanded_blk[loop-16]), 1);

    /* initialize working variables for this block */
    a = chaining_vars[0];
    b = chaining_vars[1];
    c = chaining_vars[2];
    d = chaining_vars[3];
    e = chaining_vars[4];

    /* processing stuff */
    for (loop = 0; loop < 80; loop++)
    {
        t = ROTL(a,5) + Function(loop,b,c,d) + e + expanded_blk[loop];
        e = d;
        d = c;
        c = ROTL(b,30);
        b = a;
        a = t;
    }

    /* set the hash value for next block */
    chaining_vars[0] += a;
    chaining_vars[1] += b;
    chaining_vars[2] += c;
    chaining_vars[3] += d;
    chaining_vars[4] += e;

    return;
}

/* Padding the message and calculate hash */
void sha_1_padding(unsigned char *buffer,
                   unsigned char buffer_length,
                   unsigned long long message_length)
{
    unsigned char loop;

    /* mark */
    buffer[buffer_length] = 0x80;

    /* if length < 55 there is space for message length, we process 1 block */
    if (buffer_length < 55)
    {
        for (loop = (buffer_length+1); loop < 56; loop++)
            buffer[loop] = 0x00;

        for (loop = 0; loop < 8; loop++)
            buffer[56+loop] = ((message_length) >> (56 - (8 * loop)))& 0xFF;

        sha_1_process_block(buffer);
    }
    else
    {
        /* if length > 55, must fill with zeroes and process 2 blocks */
        for (loop = (buffer_length+1); loop < 64; loop++)
            buffer[loop] = 0x00;

        sha_1_process_block(buffer);

        for (loop = 0; loop < 56; loop++ )
            buffer[loop] = 0x00;

        for (loop = 0; loop < 8; loop++)
            buffer[56+loop] = ((message_length) >> (56 - (8 * loop)))& 0xFF;

        sha_1_process_block(buffer);
    }
    return;
}

/* process the whole file using buffers of size BUFFER_SIZE */
booleantype sha_1_file_big_buffer(FILE *filename, unsigned long long filesize)
{
    unsigned long long message_length;
    unsigned long long buffer_length;
    unsigned char *buffer;

    /* if we have memory we will use this method, if not use small buffers */
    if (NULL == (buffer = calloc( 1, (BUFFER_SIZE + 64))))
        return sha_1_file_small_buffer(filename, filesize);

    /* init stage */
    sha_1_init();

    message_length = filesize;
    while (message_length >= BUFFER_SIZE)
    {
        if (BUFFER_SIZE != fread(buffer, 1, BUFFER_SIZE, filename))
            return _error_;

        buffer_length = 0;
        while (buffer_length != BUFFER_SIZE)
        {
            sha_1_process_block(&buffer[buffer_length]);
            buffer_length += 64;
        }
        message_length -= BUFFER_SIZE;
    }

    /* fill last buffer < BUFFER_SIZE */
    if (message_length != fread(buffer, 1, message_length, filename))
        return _error_;

    buffer_length = message_length;
    while (buffer_length > 64)
    {
        sha_1_process_block(&buffer[message_length-buffer_length]);
        buffer_length -= 64;
    }

    sha_1_padding(&buffer[message_length-buffer_length],buffer_length, filesize*8);

    free(buffer);
    return _success_;
}

/* process a whole file in 64 bytes blocks */
booleantype sha_1_file_small_buffer(FILE *filename, unsigned long long filesize)
{
    unsigned long long message_length;
    unsigned char file_buffer[64];

    /* init stage */
    sha_1_init();

    message_length = filesize;
    while (message_length > 64)
    {
        /* Process the file in 512 bit blocks = 64 chars */
        if (64 != fread(file_buffer, sizeof(unsigned char),64, filename))
            return _error_;

        sha_1_process_block(file_buffer);
        message_length -= 64;
    }

    /* read last block */
    if (message_length != fread(file_buffer, 1, message_length, filename))
        return _error_;

    sha_1_padding(file_buffer, message_length, filesize*8);
    return _success_;
}

/* process a memory buffer */
void sha_1_buffer(unsigned char *buffer,
                  unsigned long long message_length)
{
    unsigned long long temp_length = 0;

    /* init stage */
    sha_1_init();

    temp_length = message_length;

    /* Process the buffer in 512 bit blocks */
    while (temp_length >= 64)
    {
        sha_1_process_block(&buffer[message_length-temp_length]);
        temp_length -=64;
    }

    sha_1_padding(&buffer[message_length-temp_length],
                  temp_length,
                  message_length*8);
    return;
}


